#!/usr/bin/env python3

import click

from budget import __version__ as _version
from budget.commands.balance import balance
from budget.commands.check import check
from budget.structures import Args


def print_version(ctx, _param, value):
    if not value or ctx.resilient_parsing:
        return

    print(f"hledger-budget {_version}")
    ctx.exit()


@click.group()
@click.option("-f", "--file", help="input file")
@click.option(
    "--version", is_flag=True, callback=print_version, expose_value=False, is_eager=True
)
def cli(**kwargs):
    """Budget helper for hledger"""
    args = Args(**kwargs)
    ctx = click.get_current_context()
    ctx.obj = args


cli.add_command(balance)
cli.add_command(check)



from rich.console import Console
from rich.theme import Theme

_theme = Theme({
    "default": "",
    "error": "red",
    "list": "blue",
    "alt": "italic"
})

console = Console(emoji=False, theme=_theme)
econsole = Console(emoji=False, theme=_theme, stderr=True)



from budget.table import Table
from jinja2 import Environment, PackageLoader, select_autoescape

from ._renderer import Renderer


class JinjaRenderer(Renderer):
    def __init__(self, template: str):
        self._template = template
        self.env = Environment(
            loader=PackageLoader("budget.rendering", "templates"),
            autoescape=select_autoescape(),
        )

    def render_table(self, table: Table):
        layout = self.env.get_template(self._template)
        print(layout.render(table=table))



from datetime import date, timedelta


def first_of_this_month() -> str:
    return date.today().replace(day=1).isoformat()


def tomorrow() -> str:
    return (date.today() + timedelta(days=1)).isoformat()



import click

from budget.rendering import OutputType

_COMMON_OPTS = [
    click.option(
        "-i",
        "--budget-include",
        "includes",
        multiple=True,
        default=[],
        help="include accounts for budgeting. When not set, all accounts are included",
    ),
    click.option(
        "-x",
        "--budget-exclude",
        "excludes",
        multiple=True,
        default=[],
        help="exclude accounts for budgeting.",
    ),
    click.option(
        "-c",
        "--commodity",
        "commodities",
        multiple=True,
        default=[],
        help="include only selected currencies",
    ),
    click.option(
        "-b",
        "--begin",
        default=None,
        help="include postings since this date (default: begin of current month, if no other time period is set)",
    ),
    click.option(
        "-e",
        "--end",
        default=None,
        help="include postings until this date",
    ),
    click.option(
        "-p",
        "--period",
        default=None,
        help="time period for which report should be generated",
    ),
    click.option(
        "-O",
        "--output-type",
        type=click.Choice(tuple(ot.value for ot in OutputType)),
        default="rich",
        help="output-type",
    ),
]


def common_options(fn):
    for opt in reversed(_COMMON_OPTS):
        fn = opt(fn)
    return fn



from dataclasses import dataclass, field
from typing import Any

from rich import box
from rich.table import Table as RichTable
from rich.text import Text as RichText

from budget.table import Table, Text
from budget.utils import rprint

from ._renderer import Renderer


@dataclass
class Message:
    order: int = field(hash=False, compare=False)
    message: str

    def __hash__(self):
        return hash(self.message)


class RichRenderer(Renderer):
    def render_table(self, table: Table):
        kw: dict[str, Any] = {"box": box.MINIMAL_HEAVY_HEAD}

        if table.title:
            kw["title"] = table.title

        if any(table.header):
            kw["show_header"] = True
            kw["header_style"] = "bold"

        if table.footer:
            kw["show_footer"] = True

        rt = RichTable(**kw)
        messages: set[Message] = set()
        for i, header in enumerate(table.header):
            footer = to_rich_text(table.footer[i]) if table.footer else ""
            colstyle = table.styles[i]
            rt.add_column(to_rich_text(header), footer, justify=colstyle.justify)

        for row in table.rows:
            textrow = []
            for cell in row:
                textrow.append(to_rich_text(cell))
                msg = cell.metadata.get("message")
                if msg:
                    messages.add(Message(len(messages), msg))

            rt.add_row(*textrow)

        rprint(rt)

        if messages:
            # Use grid for pretty-printing of wrapped lines after the dash.
            # Yay, what a hack! :)
            grid = RichTable.grid(padding=(0, 1))
            grid.add_column()
            grid.add_column(style="alt")
            for m in sorted(messages, key=lambda m: m.order):
                grid.add_row("-", m.message)

            rprint(grid)


def to_rich_text(other: Text):
    return RichText(other.get_text(sep="\n"), style=other.style)



from dataclasses import dataclass, field
from decimal import Decimal
from typing import Iterable

from budget.date import first_of_this_month, tomorrow


@dataclass
class Args:
    file: str | None = None
    begin: str | None = None
    end: str | None = None
    period: str | None = None
    commodities: list[str] = field(default_factory=list)
    includes: list[str] = field(default_factory=list)
    excludes: list[str] = field(default_factory=list)
    flags: dict[str, bool | str] = field(default_factory=dict)
    output_type: str = ""

    def __post_init__(self):
        def to_list(val):
            if isinstance(val, str):
                return [str]
            if isinstance(val, tuple):
                return list(val)
            return val

        self.commodities = to_list(self.commodities)
        self.includes = to_list(self.includes)
        self.excludes = to_list(self.excludes)

        if not self.begin and not self.end and not self.period:
            self.begin = first_of_this_month()

        if not self.end and not self.period:
            # 1. hledger's --end is exclusive and we want to show today's transactions
            # 2. This has side effects: when there are no transactions, or only
            #    periodic transactions in journal file, hledger won't calculate end date
            #    and won't show the budget, because it would be infinity (hledger uses
            #    date of last transaction as default --end). This allows it to show the
            #    budget even if no transactions were made.
            self.end = tomorrow()


@dataclass
class Summary:
    account: str
    period: str
    commodity: str
    balance: Decimal | int = 0
    budget: Decimal | int = 0

    def __post_init__(self):
        if isinstance(self.balance, (str, int)):
            self.balance = Decimal(self.balance)
        if isinstance(self.budget, (str, int)):
            self.budget = Decimal(self.budget)


@dataclass
class Database:
    summaries: list[Summary] = field(default_factory=list)
    periods: list[str] = field(default_factory=list)
    commodities: list[str] = field(default_factory=list)

    def insert(self, summary: Summary):
        self.summaries.append(summary)

        if summary.period not in self.periods:
            self.periods.append(summary.period)

        if summary.commodity not in self.commodities:
            self.commodities.append(summary.commodity)

    def insert_many(self, summaries: Iterable[Summary]):
        for summary in summaries:
            self.insert(summary)

    def iter(self, **matchers):
        matchers_ = []
        for name, matchval in matchers.items():
            if matcher := self._create_matcher(name, matchval):
                matchers_.append(matcher)

        for summary in self.summaries:
            if all(m(summary) for m in matchers_):
                yield summary

    def _create_matcher(self, attr, matchval):
        if matchval is None:
            return None

        if callable(matchval):
            return matchval

        if matchval in (True, False):
            return lambda x: bool(getattr(x, attr)) is matchval

        if isinstance(matchval, (list, tuple)):
            return lambda x: getattr(x, attr) in matchval

        return lambda x: getattr(x, attr) == matchval



import csv
import os
import subprocess
from io import StringIO
from typing import Iterable

from budget.structures import Args, Database, Summary


def bal_csv(args: Args):
    cmd = ["hledger"]

    if args.file:
        cmd.extend(("--file", args.file))

    cmd.extend(["balance", "--no-total", "--layout", "bare", "-O", "csv"])

    if args.begin:
        cmd.extend(("--begin", args.begin))
    if args.end:
        cmd.extend(("--end", args.end))
    if args.period:
        cmd.extend(("--period", args.period))

    for name, value in args.flags.items():
        if value is True:
            cmd.append(f"--{name}")
        elif value is False:
            continue
        else:
            cmd.extend((f"--{name}", str(value)))
    for excl in args.excludes:
        cmd.append(f"not:{excl}")
    if args.commodities:
        expr = " or ".join(f"cur:{comm}" for comm in args.commodities)
        cmd.append(f"expr:{expr}")

    cmd.extend(args.includes)

    env = os.environ.copy()
    env.pop("LANG", None)
    cp = subprocess.run(cmd, check=True, capture_output=True, text=True, env=env)
    f = StringIO(cp.stdout)
    yield from csv.reader(f, delimiter=",")


def bal_budget(args: Args):
    args.flags["budget"] = True
    yield from bal_csv(args)


def deaggregate(summaries: Iterable[Summary]):
    revsummaries = sorted(summaries, reverse=True, key=lambda r: r.account)
    for i, summary in enumerate(revsummaries):
        for parent in revsummaries[i + 1 :]:
            if (
                summary.account.startswith(parent.account + ":")
                and summary.commodity == parent.commodity
            ):
                parent.balance -= summary.balance
                if parent.budget >= summary.budget:
                    parent.budget -= summary.budget


def prepare_db(rows, commodities_order=None) -> Database:
    if commodities_order is None:
        commodities_order = []

    db = Database(commodities=list(commodities_order))
    header = next(rows)

    for row in rows:
        # csv format: account, commodity, p1bal, [p1bud], p2bal, [p2bud], ...
        account = row[0]
        commodity = row[1]

        i = 2
        while i < len(row):
            period = header[i]
            balance = row[i]
            budget = 0

            if i + 1 < len(row) and header[i + 1] == "budget":
                budget = row[i + 1]
                i += 1

            i += 1
            s = Summary(account, period, commodity, balance, budget)

            # hledger shows previously budgeted accounts as empty, but doesn't
            # include their commodity when there's no budget for them in
            # current time period.
            if not s.commodity and not s.balance and not s.budget:
                continue

            db.insert(s)

    if commodities_order is None:
        commodities_order = []

    for period in db.periods:
        deaggregate(db.iter(period=period))

    return db



import sys
from dataclasses import replace

import click
from budget.hledger import bal_budget, bal_csv, prepare_db
from budget.rendering import render
from budget.structures import Args, Database, Summary
from budget.table import ColumnStyle, Table, Text
from budget.utils import eprint

from ._options import common_options


def sum_budgeted(db: Database, commodity: str) -> Summary:
    summed = Summary("total", "", commodity, 0, 0)
    for summary in db.iter(commodity=commodity):
        if not summary.budget:
            continue

        summed.balance += summary.balance
        if summed.budget is None:
            summed.budget = summary.budget
        else:
            summed.budget += summary.budget
    return summed


def sum_missing_spendings(db: Database, commodity: str) -> Summary:
    fore = Summary("forecast", "", commodity, 0, 0)

    for summary in db.iter(commodity=commodity):
        if not summary.budget:
            continue

        if summary.budget < summary.balance:
            continue

        fore.balance += summary.budget - summary.balance

    return fore


@click.command()
@common_options
@click.argument("assets", nargs=-1)
@click.option(
    "-t",
    "--title",
    default="Budget Verification",
    help="report title",
)
@click.option(
    "--no-auto-excludes",
    is_flag=True,
    default=False,
    help="disable automatic exclusions of unbudgeted and assets accounts from the budget",
)
@click.pass_obj
def check(args: Args, title, assets, no_auto_excludes, **kwargs):
    """Check correctness of budget against available assets."""
    if not assets:
        return

    args = replace(args, **kwargs)

    # This is what users actually want even if they don't know it :)
    #
    # Suppose that we have 2 accounts for our cash:
    #   - assets:bank:checking
    #   - assets:bank:savings
    # Each month we save some cash by transfering it from checking to savings.
    # Hledger includes savings account in budget report, but it also
    # includes its parent: assets:bank. Now, assets:bank:checking quietly
    # contributes to assets:bank (see `hledger bal -E` for proof) and to
    # the sum of expenses. The problem is that checking accounts balance is a
    # result of many transactions which don't contribute to the "budget" or
    # "spending" report, like income.
    if not no_auto_excludes:
        args.excludes.extend(assets)
        args.excludes.append("unbudgeted")

    db = prepare_db(bal_budget(args))

    ass_args = replace(
        args,
        includes=assets,
        excludes=[],
        flags={"cumulative": True, "historical": True},
    )

    assets_db = prepare_db(bal_csv(ass_args))

    result = True

    def make_cell(msg, cond: bool, errmsg: str | None = None):
        nonlocal result
        result &= cond

        t = Text(str(msg))
        if not cond:
            t.style = "error"
            if errmsg:
                t.metadata["message"] = errmsg
        return t

    if not db.commodities:
        eprint("No postings to check!")
        sys.exit(1)

    t = Table(title=title)
    t.add_column(
        "Money available",
        "Remaining budgeted spendings",
        "Budgeted spendings",
        "Total forecasted spendings",
        "Total budget",
        "Unbudgeted funds",
        header="Report",
    )

    for commodity in db.commodities:
        spendings = sum_budgeted(db, commodity)
        if spendings.budget is None:
            eprint(f"No budgeted spendings for {commodity}.")
            eprint("Do you have any periodic transactions for this commodity?")
            result = False
            continue

        missing_spendings = sum_missing_spendings(db, commodity)
        forecast = spendings.balance + missing_spendings.balance

        # NOTE: in case there's no assets in a given currency, assets_db.iter()
        # will produce empty iterator, and sum() will return integer (not Decimal) 0.
        # This is the case when you budget in a currency, but rely on auto
        # currency conversion done by bank.
        cash = sum(summary.balance for summary in assets_db.iter(commodity=commodity))

        unbudgeted = cash - missing_spendings.balance

        t.add_column(
            f"{cash:.2f}",
            make_cell(
                f"{missing_spendings.balance:.2f}",
                missing_spendings.balance <= cash,
                "Not enough money for remaining spendings.",
            ),
            f"{spendings.balance:.2f}",
            make_cell(
                f"{forecast:.2f}",
                forecast <= cash + spendings.balance,
                "Insufficient funds for forecasted budgeted and unbudgeted spendings.",
            ),
            f"{spendings.budget:.2f}",
            make_cell(f"{unbudgeted:.2f}", unbudgeted >= 0, "Over-budgeting."),
            header=commodity,
            style=ColumnStyle(justify="right"),
        )

    render(args.output_type, t)

    sys.exit(int(not result))



from dataclasses import replace
from typing import Any, Iterable

import click

from budget.hledger import bal_budget, prepare_db
from budget.rendering import render
from budget.structures import Args, Summary
from budget.table import ColumnStyle, Table
from budget.utils import flatten

from ._options import common_options


def add_to_row(
    s: Summary, commodities: list[str], rows: dict[str, dict[str, Summary | None]]
):
    if s.account not in rows:
        rows[s.account] = {comm: None for comm in commodities}
    rows[s.account][s.commodity] = s


def collect_rows(
    summaries: Iterable[Summary], commodities: list[str]
) -> dict[str, dict[str, Summary | None]]:
    rows: dict[str, dict[str, Summary | None]] = {}
    for summary in summaries:
        add_to_row(summary, commodities, rows)
    return rows


def to_cells(s: Summary | None) -> tuple[str, str]:
    if not s:
        return "", ""

    if not s.budget:
        return f"{s.balance:.2f}", ""

    perc = round(s.balance / s.budget * 100)
    return f"{s.balance:.2f}", f"{perc}% of {s.budget:.2f}"


@click.command()
@common_options
@click.option("-t", "--title", default="Budget: {commodity}", help="report title")
@click.option(
    "--no-auto-excludes",
    is_flag=True,
    default=False,
    help="disable automatic exclusions of unbudgeted accounts",
)
@click.option(
    "-u",
    "--unbudgeted",
    is_flag=True,
    default=False,
    help="show accounts without budget goals",
)
@click.option(
    "-U",
    "--only-unbudgeted",
    is_flag=True,
    default=False,
    help="show only accounts without budget goals; implies --unbudgeted",
)
@click.option(
    "-E",
    "--empty",
    is_flag=True,
    default=False,
    help="show accounts which have no budget or balance",
)
@click.option(
    "--no-total",
    is_flag=True,
    default=False,
    help="don't show a summary of budget and spendings",
)
@click.pass_obj
def balance(
    args: Args,
    title,
    no_auto_excludes,
    unbudgeted,
    only_unbudgeted,
    empty,
    no_total,
    **kwargs,
):
    """Show balance report together with budget goals defined by periodic transactions.

    This report differs from ordinary 'hledger balance --budget' by showing
    only accounts with budget goals without aggregating goals of subaccounts.

    When run with --unbudgeted, report will also present accounts without
    budget goals. Side effect is that due to the presence of unbudgeted
    accounts, balances for some of budgeted parent accounts will change to 0.

    Report is created for all accounts, but you should probably exclude all
    accounts which are transaction sources (typically "assets:bank:checking"
    and so on).
    """
    if only_unbudgeted:
        unbudgeted = True

    args = replace(args, **kwargs)

    # This is counter-intuitive: why we use the value of --unbudgeted instead
    # of --empty? That's because 'hledger --budget' hides subaccounts without
    # budget, so if we want to show unbudgeted accounts, we must use 'hledger
    # --empty'.
    #
    # Side effect is that our de-aggregation now has a complete list of
    # accounts, so main parents will always have 0 balance, even with
    # additional exclusions (because hledger will take exclusions into account
    # and re-calculate parents' balances)
    args.flags["empty"] = unbudgeted

    if not no_auto_excludes:
        args.excludes.append("unbudgeted")

    db = prepare_db(bal_budget(args), commodities_order=kwargs["commodities"])

    matchers: dict[str, Any] = {}
    if not unbudgeted:
        matchers["budget"] = True
    if only_unbudgeted:
        matchers["budget"] = False
    if not empty:
        matchers["_"] = lambda x: x.balance or x.budget

    summaries = collect_rows(db.iter(**matchers), db.commodities)
    total: list[Summary] = [Summary("Total", "", c) for c in db.commodities]

    t = Table(title=title.format(commodity=", ".join(db.commodities)))
    t.add_column(header="Account", *summaries.keys())
    for i, comm in enumerate(db.commodities):
        cbal = []
        cexe = []

        for summary_per_comm in summaries.values():
            summary = summary_per_comm[comm]
            bal, execution = to_cells(summary)
            cbal.append(bal)
            cexe.append(execution)
            if summary:
                total[i].balance += summary.balance
                total[i].budget += summary.budget

        t.add_column(
            *cbal, header=["Balance", comm], style=ColumnStyle(justify="right")
        )
        t.add_column(
            *cexe, header=["Budget Execution", comm], style=ColumnStyle(justify="right")
        )

    if not no_total:
        footer = flatten([to_cells(t) for t in total])
        t.set_footer("Total", *footer)

    render(args.output_type, t)



import csv
import sys

from budget.table import Table

from ._renderer import Renderer


class CSVRenderer(Renderer):
    def render_table(self, table: Table):
        writer = csv.writer(sys.stdout)

        if table.header:
            writer.writerow(h.text for h in table.header)

        for row in table.rows:
            writer.writerow([cell.text for cell in row])

        if table.footer:
            writer.writerow([f.text for f in table.footer])

        sys.stdout.flush()



from abc import ABCMeta, abstractmethod
from functools import singledispatchmethod

from budget.table import Table


class Renderer(metaclass=ABCMeta):
    @singledispatchmethod
    def render(self, arg):
        raise TypeError("unknown renderable type")

    @render.register
    def _(self, table: Table):
        return self.render_table(table)

    @abstractmethod
    def render_table(self, table: Table):
        pass



from dataclasses import dataclass, field
from typing import Any, Literal


@dataclass
class Text:
    lines: str | list[str]
    style: str = "default"
    metadata: dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        if not isinstance(self.lines, (str, list, tuple)):
            self.lines = str(self.lines)

    @classmethod
    def from_any(cls, lhs) -> "Text":
        if not isinstance(lhs, Text):
            return Text(lhs)
        return lhs

    @property
    def text(self) -> str:
        return self.get_text()

    def get_text(self, sep: str = " ") -> str:
        if isinstance(self.lines, str):
            return self.lines
        return sep.join(self.lines)


@dataclass
class ColumnStyle:
    justify: Literal["default", "left", "center", "right", "full"] = "left"


class Table:
    def __init__(self, title=None):
        self.title: str | None = title
        self.header: list[Text] = []
        self.rows: list[list[Text]] = []
        self.footer: tuple[Text, ...] = ()
        self.styles: list[ColumnStyle] = []

    def add_column(
        self,
        *cells,
        header: Text | str | list[str] = "",
        style: ColumnStyle | None = None
    ):
        if self.rows and len(cells) != len(self.rows):
            raise ValueError("number of items doesn't match number of rows")
        if not self.rows:
            self.rows = [[] for _ in cells]

        if style is None:
            style = ColumnStyle()

        for i, cell in enumerate(cells):
            self.rows[i].append(Text.from_any(cell))

        self.header.append(Text.from_any(header))
        self.styles.append(style)

    def add_row(self, *cells: str | Text):
        if self.rows and len(self.rows[0]) != len(cells):
            raise ValueError("number of items doesn't match number of columns")

        self.rows.append([Text.from_any(cell) for cell in cells])

    def set_footer(self, *data):
        if len(data) != len(self.header):
            raise ValueError("invalid number of items in footer")
        self.footer = tuple(Text.from_any(d) for d in data)



import itertools
from dataclasses import dataclass
from decimal import Decimal

from budget.console import console, econsole

rprint = console.print


@dataclass
class Amount:
    balance: Decimal = Decimal(0)
    budget: Decimal = Decimal(0)


def eprint(*a, **kw):
    kw.setdefault("style", "error")
    econsole.print(*a, **kw)


def flatten(val):
    return list(itertools.chain.from_iterable(val))



from enum import Enum

from ._renderer import Renderer
from .csv import CSVRenderer
from .jinja import JinjaRenderer
from .rich import RichRenderer


class OutputType(Enum):
    rich = "rich"
    csv = "csv"
    html = "html"
    html_bare = "html-bare"
    dokuwiki = "dokuwiki"


def render(output_type, *objects):
    renderer: Renderer | None = None
    match OutputType(output_type):
        case OutputType.rich:
            renderer = RichRenderer()
        case OutputType.csv:
            renderer = CSVRenderer()
        case OutputType.html:
            renderer = JinjaRenderer("html_full.html")
        case OutputType.html_bare:
            renderer = JinjaRenderer("html_bare.html")
        case OutputType.dokuwiki:
            renderer = JinjaRenderer("dokuwiki.txt")
        case _:
            raise ValueError(f"unknown renderer: {output_type}")

    for obj in objects:
        renderer.render(obj)
